# -*- coding: utf-8 -*-
import logging

from logging import handlers
from malibu.util import get_caller


class LoggingDriver(object):

    __instances = {}

    @classmethod
    def get_instance(cls, name=None):

        if name is None:
            name = get_caller().split('.')[0]

        if not cls.__instances or name not in cls.__instances:
            return None

        return cls.__instances.get(name)

    @classmethod
    def find_logger(cls, name=None):

        if name is None:
            name = get_caller()
            root = name.split('.')[0]
        else:
            root = name

        if not cls.get_instance(name=root):
            return None

        return cls.get_instance(name=root).get_logger(name=name)

    @classmethod
    def from_config(cls, config, name=None):

        if not name:
            name = get_caller().split('.')[0]
        else:
            name = name

        logfile = config.get_string("logfile",
                                    "/var/log/{}.log".format(
                                        name))
        loglevel = config.get_string("loglevel", "INFO").upper()
        stream = config.get_bool("console_log", True)

        loglevel = getattr(logging, loglevel, None)
        if not isinstance(loglevel, int):
            raise TypeError("Invalid log level: {}".format(
                config.get_string("loglevel", "INFO").upper()))

        return cls(logfile=logfile, loglevel=loglevel,
                   stream=stream, name=name)

    def __init__(self, logfile, loglevel, stream, name=None):
        """ __init__(self, name=None)

            Initializes the logging driver and loads necessary config
            values from the ConfigurationSection that should be passed
            in as config. Also loads the root logger from a specified
            default name or from the base module name of the package.
        """

        if not name:
            self.name = get_caller().split('.')[0]
        else:
            self.name = name

        self.__logfile = logfile
        self.__stream = stream

        if isinstance(loglevel, int):
            self.__loglevel = loglevel
        else:
            self.__loglevel = getattr(logging, loglevel, None)
            if not isinstance(self.__loglevel, int):
                raise TypeError("Invalid log level: {}".format(loglevel))

        LoggingDriver.__instances.update({self.name: self})

        self.__setup_logger()

    def __setup_logger(self):
        """ __setup_logger(self)

            Sets up the logging system with the logfile, loglevel, and
            other streaming options.
        """

        # Set up logging with the root handler so all log objects get the same
        # configuration.
        logger = logging.getLogger(self.name)
        logger.setLevel(self.__loglevel)
        formatter = logging.Formatter(
            '%(asctime)s | %(name)-50s %(levelname)s : %(message)s')

        file_logger = handlers.RotatingFileHandler(
            self.__logfile,
            maxBytes=8388608,  # 8MB
            backupCount=4)
        file_logger.setLevel(self.__loglevel)
        file_logger.setFormatter(formatter)

        logger.addHandler(file_logger)

        if self.__stream:
            logger.debug(" --> Building streaming log handler...")
            stream_logger = logging.StreamHandler()
            stream_logger.setLevel(self.__loglevel)
            stream_logger.setFormatter(formatter)

            logger.addHandler(stream_logger)

        logger.info(" --> Logfile is opened at: {}".format(self.__logfile))
        logger.info(" --> Console / streaming logs: {}".format(
            "enabled" if self.__stream else "disabled"))
        logger.info(" --> Logging at level: {}".format(self.__loglevel))

    def get_logger(self, name=None):
        """ get_logger(self, name=None)

            Will return a logger object for a specific namespace.
            If name parameter is None, get_logger will use call
            stack inspection to get the namespace of the last caller.
        """

        if name is None:
            name = get_caller()

        return logging.getLogger(name)



# -*- coding: utf-8 -*-
# ASCII / ANSI Escape codes.

# Graphics mode / text styles.
STYLE_OFF = '\033[0m'
STYLE_BOLD = '\033[1m'
STYLE_UNDERSCORE = '\033[4m'
STYLE_BLINK = '\033[5m'
STYLE_REVERSE_VID = '\033[7m'
STYLE_CONCEALED = '\033[8m'

# Foreground text colors.
FG_BLACK = '\033[0;30m'
FG_RED = '\033[0;31m'
FG_GREEN = '\033[0;32m'
FG_BROWN = '\033[0;33m'
FG_BLUE = '\033[0;34m'
FG_MAGENTA = '\033[0;35m'
FG_CYAN = '\033[0;36m'
FG_LGRAY = '\033[0;37m'

FG_DGRAY = '\033[1;30m'
FG_LRED = '\033[1;31m'
FG_LGREEN = '\033[1;32m'
FG_YELLOW = '\033[1;33m'
FG_LBLUE = '\033[1;34m'
FG_LPURPLE = '\033[1;35m'
FG_LCYAN = '\033[1;36m'
FG_WHITE = '\033[1;37m'

# Background text colors.
BG_BLACK = '\033[0;40m'
BG_RED = '\033[0;41m'
BG_GREEN = '\033[0;42m'
BG_BROWN = '\033[0;43m'
BG_BLUE = '\033[0;44m'
BG_MAGENTA = '\033[0;45m'
BG_CYAN = '\033[0;46m'
BG_LGRAY = '\033[0;47m'

BG_DGRAY = '\033[1;40m'
BG_LRED = '\033[1;41m'
BG_LGREEN = '\033[1;42m'
BG_YELLOW = '\033[1;43m'
BG_LBLUE = '\033[1;44m'
BG_LPURPLE = '\033[1;45m'
BG_LCYAN = '\033[1;46m'
BG_WHITE = '\033[1;47m'


def style_text(style, obj):
    """ Simple method to stringify (or coerce an object to
        a string type, for those who prefer political correctness)
        with ASCII color or style codes and automatically terminate
        with STYLE_OFF escapes.
    """

    return style + str(obj) + STYLE_OFF



# -*- coding: utf-8 -*-
import json
import sqlite3


def install_json_converter():
    """ Installs a json object converter into the sqlite3 module for
        quick type conversions.
    """

    sqlite3.register_converter("json", __convert_json)


def __convert_json(string):
    """ Converts a string into JSON format.  If the conversion fails,
        returns None.
    """

    try:
        return json.loads(string)
    except:
        return None



# -*- coding: utf-8 -*-
import traceback
from malibu.design import borgish


__doc__ = """
malibu.command.module
---------------------

A relatively self-contained system for loading and creating command "modules"
for a CLI-style script.

A command module must extend the :class:`~malibu.command.module.CommandModule`
and set up the class as such:

.. code-block:: python

    from malibu.command import command_module, module

    @command_module(
        name = "example",
        depends = []
    )
    class ExampleModule(module.CommandModule):

        BASE = "example"

        def __init__(self, loader):

            super(ExampleModule, self).__init__(base = ExampleModule.BASE)
            self.__loader = loader

            self.register_subcommand("help", self.show_help)

        def show_help(self, *args, **kw):
            \"\"\" example:help []

                Does something.
            \"\"\"

            if "args" in kw:
                argparser = kw["args"]
            else:
                argparser = self.__loader.get_argument_parser()

            pass  # Do stuff...


After all modules are implemented, use something similar to the following in
a console entry point:

.. code-block:: python

    from malibu import command
    from malibu.util import args

    argparser = args.ArgumentParser.from_argv()
    # Set up argparser params, mappings, etc here.

    modloader = module.CommandModuleLoader(argparser_

    mods = command.get_command_modules(package = __package__)
    # Or replace __package__ with your cmd module package path
    modloader.register_modules(mods.values())
    modloader.instantiate_modules()

    argparser.parse()

    modloader.parse_command(
        argparser.parameters[1],  # module base
        *argparser.parameters[1:],  # subcommand and args
        args = argparser)

.. autoclass:: CommandModuleLoader
   :members:

.. autoclass:: CommandModule
   :members:

.. autoexception:: CommandModuleException
   :members:
"""


class CommandModuleLoader(borgish.SharedState):

    def __init__(self, argparser, *args, **kw):
        """ Initializes a ModuleLoader object with a list for modules and
            sets the static __instance so the object is always accessible.
        """

        super(CommandModuleLoader, self).__init__(*args, **kw)

        if "state" not in kw:
            self.__modules = []
            self.__modclasses = []

            self.__argparser = argparser

    def register_module(self, cls):
        """ Registers a single module in the modloader list.

            Checks if any module instances in the module list
            and raises if an instance already exists.

            Otherwise, the module is instantiated, appended,
            and returned.
        """

        instances = filter(lambda mod: isinstance(mod, cls), self.__modules)
        if True in instances:
            raise CommandModuleException("Cannot re-register module %s" %
                                         (cls.__name__))

        if cls in self.__modclasses:
            raise CommandModuleException("Cannot re-register module %s" %
                                         (cls.__name__))

        self.__modclasses.append(cls)

    def register_modules(self, clslist):
        """ Registers a list of modules through subsequent calls to
            register_module().  Returns the list of instantiated
            modules.
        """

        [self.register_module(module) for module in clslist]

    def instantiate_modules(self, clslist=[]):
        """ Instantiates all module classes that are registered.
            ** Might perform dependency lookup as well.
        """

        deferred = []

        if clslist == []:
            clslist = self.__modclasses

        for cls in clslist:
            for dep in cls.depend_modules:
                if self.get_module_by_base(dep) is None:
                    deferred.append(cls)
                    break
            if cls in deferred:
                continue
            else:
                self.__modules.append(cls(self))

        if len(deferred) != 0:
            self.instantiate_modules(deferred)

    def deinit_modules(self):
        """ Runs __deinit__ on all registered modules.
        """

        for module in self.__modules:
            module.__deinit__()

    def get_argument_parser(self):
        """ Returns the argument parser that will be passed into
            functions that are called by the loader as command line
            parameters.
            Allows modules to access the parser during instantiation
            to change param modules, add help text, register aliases,
            etc.
        """

        return self.__argparser

    def get_module_by_base(self, modbase):
        """ Returns a module instance by the module's base name.
            Returns None if the named instance does not exist.
        """

        for module in self.__modules:
            if module.get_base().lower() == modbase.lower():
                return module

        return None

    @property
    def modules(self):
        """ Returns the list of modules.
        """

        return self.__modules

    def deregister_module(self, obj):
        """ Removes a module instance from the list of registered
            modules.
        """

        if obj not in self.__modules:
            return None

        return self.__modules.remove(obj)

    def parse_command(self, command, *args, **kw):
        """ Process a command and fire the function for the matching
            command and subcommand.  Returns the function execution
            result, if any.
        """

        try:
            command, subcommand = command.split(':')
        except:
            command, subcommand = ['help', 'show']
        for module in self.__modules:
            if not module.is_command(command):
                continue
            if (not module.has_subcommand(subcommand) and not
                    module.has_alias(subcommand)):

                continue
            try:
                result = module.execute_subcommand(subcommand, *args, **kw)
                return result
            except:
                traceback.print_exc()
                raise CommandModuleException(
                    "Exception occurred while executing a command: "
                    "[command=%s subcommand=%s args=%s kw=%s]" % (
                        command, subcommand, args, kw))


class CommandModule(object):
    """ Module superclass.  Abstracts away some parts of
        a command module to make implementation simpler.
        Should only be inherited, never instantiated by itself.
    """

    def __init__(self, base=None):
        """ Initializes a Module object with the command base,
            maps, and help dictionary.
        """

        self.__command_base = base

        self.__command_map = {}
        self.__command_alias_map = {}

        self.__command_help = {}

    def __deinit__(self):
        """ Used during "system" shutdown.  Perform clean up of
            module data and close any open parts of the system
            that should be properly deinitialized.
        """

        pass

    def is_command(self, command):
        """ Simple boolean-returning method used during command
            parsing.
        """

        return command == self.__command_base

    def has_subcommand(self, subcommand):
        """ Boolean-returning method for if this Module
            has registered a specific subcommand.
        """

        return subcommand in self.__command_map.keys()

    def has_alias(self, alias):
        """ Boolean-returning method for if this Module
            has registered a specific alias.
        """

        return alias in self.__command_alias_map.keys()

    def resolve_alias(self, alias):
        """ Resolves an alias to a subcommand and returns
            the subcommand.
        """

        try:
            return self.__command_alias_map[alias]
        except:
            return None

    def get_base(self):
        """ Returns the command base.
        """

        return self.__command_base

    def get_help(self):
        """ Returns the help dictionary for this
            module.
        """

        return self.__command_help

    def register_subcommand(self, subcommand, function, aliases=[]):
        """ Registers a subcommand and its help in
            the internal maps. Updates aliases, subcommands, etc.
        """

        if subcommand in self.__command_map:
            raise CommandModuleException("Subcommand is already registered")
        else:
            self.__command_map.update({subcommand: function})
            if not function.__doc__:
                self.__command_help.update({subcommand: None})
            else:
                self.__command_help.update({
                    subcommand:
                    '\n'.join(
                        [line.strip() for line in function.__doc__.splitlines()
                         if line.strip() is not '']
                    )
                })

        for alias in aliases:
            if alias in self.__command_alias_map:
                raise CommandModuleException(
                    "Alias is already registered for subcommand %s" %
                    (subcommand))
            else:
                self.__command_alias_map.update({alias: subcommand})

    def unregister_subcommand(self, subcommand):
        """ Removes a subcommand, all aliases, and all help
            from the internal maps.
        """

        if subcommand not in self.__command_map:
            raise CommandModuleException(
                "Tried to remove nonexistent subcommand %s" %
                (subcommand))

        self.__command_map.pop(subcommand)

        remove_aliases = []
        for alias, sub in self.__command_alias_map.items():
            if sub == subcommand:
                remove_aliases.append(alias)

        [self.__command_alias_map.pop(alias) for alias in remove_aliases]

    def execute_subcommand(self, subcommand, *args, **kw):
        """ Attempts to fire the subcommand with arguments,
            and keywords, may throw CommandModuleException.
        """

        if subcommand not in self.__command_map:
            raise CommandModuleException(
                "Tried to execute unknown subcommand %s" %
                (subcommand))

        func = self.__command_map[subcommand]
        func(*args, **kw)


class CommandModuleException(Exception):
    """ Super special exception for modules.
    """

    def __init__(self, value):

        Exception.__init__(self)

        self.value = value

    def __str__(self):

        return repr(self.value)



# -*- coding: utf-8 -*-
import glob
import os
from importlib import import_module
from malibu.util.decorators import function_kw_reg

modules = glob.glob(os.path.dirname(__file__) + "/*.py")
__all__ = [os.path.basename(f)[:-3] for f in modules
           if not os.path.basename(f).startswith('_') and
           not f.endswith('__init__.py') and os.path.isfile(f)]


__doc__ = """
Module for processing commands in a CLI fashion.

.. autofunction:: get_command_modules
.. py:function:: command_module(func, *args, **kw)

   A decorator function that is used to register command modules in the
   :py:data:`~__command_modules` dictionary.

   :param function func: Function being decorator
   :param *args: positional arguments
   :param **kw: keyword arguments
   :type *args: list
   :type **kw: dict
   :return: none
   :rtype: None

.. automodule:: malibu.command.module
"""

__command_modules = {}
command_module = function_kw_reg(__command_modules, ["name", "depends"])


def get_command_modules(package=None):
    """ Reads a package and returns a dictionary of modules decorated with
        the :py:func:`command_module` decorator.

        :param str package: Package to search for command modules
        :return: dictionary of command modules
        :rtype: dict
        :raises AttributeError: if package has no __all__ attribute
    """

    package = __package__ if not package else package
    package_all = import_module(package)
    if not hasattr(package_all, "__all__"):
        raise AttributeError("Package %s has no __all__ attribute" % (package))

    package_all = package_all.__all__

    modules = {}
    deps = set()

    for module in package_all:
        module = import_module("{}.{}".format(package, module))
        # Just importing the code should take care of registration with the
        # decorator.

    for module, kws in __command_modules.items():
        module.depend_modules = kws["depends"]
        for depmod in kws["depends"]:
            deps.add(depmod)
        if kws["name"] in modules:
            # Module is already in map, don't clobber
            continue
        modules.update({kws["name"]: module})

    for module, kws in __command_modules.items():
        for depmod in kws["depends"]:
            if depmod not in modules:
                modules.pop(kws["name"])

    return modules



# -*- coding: utf-8 -*-
from os.path import dirname, expanduser, normpath, realpath


def expand_path(path):
    """ Expands a path to it's full, non-relative, fully
        resolved counterpart.
    """

    return realpath(normpath(expanduser(path)))


def get_path_base(path):
    """ Returns the parent of the specified path. Fully resolves
        the path before getting the basename. Uses expand_path()
        for expansion.
    """

    return dirname(expand_path(path))



# -*- coding: utf-8 -*-
import copy
import re
import sqlite3
import types
from malibu.database import dbtypeconv


class DBMapper(object):
    """ This is code for a relatively small ORM for SQLite built
        on top of the python-sqlite3 module.

        Note from the author: (01 / 14 / 2016)

        I've got to be honest, this is probably the worst code I have ever written and read.
        At this point, this code is so difficult to maintain and keep up to date for 2/3 compat that
            it is almost not worth the work.
        Especially considering that there are things like Peewee, SQLAlchemy, etc, this is not worth
            using or maintaining.

        From this point forward, I recommend using some other, cleaner, better maintained solution
            such as Peewee.
        This DBMapper code will no longer be maintained and will be deprecated starting
            with the 0.1.6 release.
        The code will be removed as the 1.0.0 release approaches.
        There may be plans to replace this with a SQLite adapter for the malibu.design.brine series
            of classes that behave similar to this, just without all the cruft.
    """

    # FETCH Constants for __execute()

    FETCH_ONE = 'one'
    FETCH_MANY = 'many'
    FETCH_ALL = 'all'

    # INDEX Constants for options dictionary.

    INDEX_PRIMARY = 'primaryIndex'
    INDEX_AUTOINCR = 'autoincrIndex'
    INDEX_UNIQUE = 'uniqueIndices'
    GENERATE_FTS_VT = 'genFTSVTs'

    # Global variables for static database methods

    _options = None
    __default_options = {
        INDEX_PRIMARY: 0,
        INDEX_AUTOINCR: True,
        INDEX_UNIQUE: set(),
        GENERATE_FTS_VT: False  # Do NOT generate FTS by default.
    }

    @staticmethod
    def get_default_options():
        """ DBMapper.get_default_options()

            Returns a deep copy of the default options dictionary for
            modification in subclasses.
        """

        return copy.deepcopy(DBMapper.__default_options)

    @staticmethod
    def connect_database(dbpath):
        """ DBMapper.connect_database(dbpath)

            Connects to a database at 'dbpath' and installs the json
            type converter "middleware" into the database system.
        """

        dbtypeconv.install_json_converter()
        __db = sqlite3.connect(dbpath, detect_types=sqlite3.PARSE_DECLTYPES)

        return __db

    @classmethod
    def set_db_options(cls, db, keys, ktypes, options=__default_options):
        """ DBMapper.set_db_options(db       => database instance
                                    keys     => list of keys
                                    ktypes   => list of key types
                                    options  => options dictionary (optional))

            Sets options for a subclasses DBMapper context.
        """

        if cls._options is None:
            cls._options = {}
            cls._options['database'] = db
            cls._options['keys'] = keys
            cls._options['keytypes'] = ktypes
            cls._options['options'] = options
        else:
            cls._options['database'] = db
            cls._options['keys'] = keys
            cls._options['keytypes'] = ktypes

    @classmethod
    def load(cls, **kw):
        """ DBMapper.load(**kw)

            Loads a *single* row from the database and populates it into
            the context cls this method was called under.

            If the database returns more than one row for the kwarg query,
            this method will only return the first result! If you want a
            list of matching rows, use find() or search().
        """

        if cls._options is None:
            raise DBMapperException(
                'Static database options have not been set.')

        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()

        keys = []
        vals = []
        for key, val in kw.items():
            keys.append(key)
            vals.append(val)
        whc = []
        for pair in zip(keys, vals):
            whc.append("%s=?" % (pair[0]))
        query = "select * from %s where (%s)" % (obj._table, ' and '.join(whc))
        result = obj.__execute(cur, query, args=vals)
        if result is None:
            for key in dbo['keys']:
                setattr(obj, "_%s" % (key), None)
            return
        for key, dbv in zip(dbo['keys'], result):
            setattr(obj, "_%s" % (key), dbv)

        return obj

    @classmethod
    def new(cls, **kw):
        """ DBMapper.new(**kw)

            Creates a new contextual instance and returns the object.
            Only parameters defined in the kwargs will be passed in to
            the record creation query, as there is no support for default
            values yet. (06/11/15)
        """

        if cls._options is None:
            raise DBMapperException(
                'Static database options have not been set.')

        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()

        keys = []
        vals = []
        for key, val in kw.items():
            keys.append(key)
            vals.append(val)
        anonvals = []
        for val in vals:
            anonvals.append('?')
        query = "insert into %s (%s) values (%s)" % (
            obj._table, ','.join(keys), ','.join(anonvals))
        obj.__execute(cur, query, args=vals)

        res = cls.find(**kw)
        if len(res) == 0:
            return None
        else:
            return res[0]

    @classmethod
    def find(cls, **kw):
        """ DBMapper.find(**kw)

            Searches for a set of records that match the query built by
            the contents of the kwargs and returns a filterable list of
            contextualized results that can be modified.
        """

        if cls._options is None:
            raise DBMapperException(
                'Static database options have not been set.')

        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()
        primaryKey = dbo['keys'][dbo['options'][DBMapper.INDEX_PRIMARY]]

        keys = []
        vals = []
        for key, val in kw.items():
            keys.append(key)
            vals.append(val)
        whc = []
        for pair in zip(keys, vals):
            whc.append('%s=?' % (pair[0]))
        query = "select %s from %s where (%s)" % (
            primaryKey, obj._table, ' and '.join(whc))
        result = obj.__execute(cur, query, args=vals, fetch=DBMapper.FETCH_ALL)

        load_pairs = []
        for row in result:
            load_pairs.append(
                {primaryKey: row[dbo['options'][DBMapper.INDEX_PRIMARY]]}
            )

        return DBResultList([cls.load(**pair) for pair in load_pairs])

    @classmethod
    def find_all(cls):
        """ DBMapper.find_all()

            Finds all rows that belong to a table and returns a filterable
            list of contextualized results. Please note that the list that
            is returned can be empty, but it should never be none.
        """

        if cls._options is None:
            raise DBMapperException(
                'Static database options have not been set.')

        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()
        primaryKey = dbo['keys'][dbo['options'][DBMapper.INDEX_PRIMARY]]

        query = "select %s from %s" % (primaryKey, obj._table)
        result = obj.__execute(cur, query, fetch=DBMapper.FETCH_ALL)

        load_pairs = []
        for row in result:
            load_pairs.append(
                {primaryKey: row[dbo['options'][DBMapper.INDEX_PRIMARY]]}
            )

        return DBResultList([cls.load(**pair) for pair in load_pairs])

    @classmethod
    def search(cls, param):
        """ DBMapper.search(param)

            This function will return a list of results that match the given
            param for a full text query. The search parameter should be in the
            form of a sqlite full text query, as defined here:
              http://www.sqlite.org/fts3.html#section_3

            As an example, suppose your table looked like this:

              +----+---------+----------------+
              | id |   name  |    description |
              +----+---------+----------------+
              | 1  |  linux  |   some magic   |
              | 2  | freebsd | daemonic magic |
              | 3  | windows |   tomfoolery   |
              +----+---------+----------------+

            A full text query for "name:linux magic" would return the first
            row because the name is linux and the description contains "magic".
            A full text query just for "description:magic" would return both
            rows one and two because the descriptions contain the word "magic".
        """

        if cls._options is None:
            raise DBMapperException(
                'Static database options have not been set.')

        if not cls._options['options'][DBMapper.GENERATE_FTS_VT]:
            raise DBMapperException(
                'Full-text search table not enabled on this table.')

        dbo = cls._options
        obj = cls(dbo['database'])
        cur = dbo['database'].cursor()

        query = """select docid from _search_%s where _search_%s match \"?\"""" % \
                (obj._table, obj._table)

        result = obj.__execute(
            cur,
            query,
            args=[param],
            fetch=DBMapper.FETCH_ALL)

        load_pairs = []
        for row in result:
            load_pairs.append({cls._options['keys'][0]: row[0]})

        return DBResultList([cls.load(**pair) for pair in load_pairs])

    @classmethod
    def join(cls, cond, a, b):
        """ DBMapper.join(cond => other table to join on
                          a    => left column to join
                          b    => right column to join)

            Performs a sqlite join on two tables. Returns the join results
            in a filterable list.
        """

        if cls._options is None or cond._options is None:
            raise DBMapperException(
                'Static database options have not been set.')

        dba = cls._options
        obja = cls(dba['database'])
        dbb = cond._options
        objb = cond(dbb['database'])
        cur = dba['database'].cursor()
        primaryKeyA = dba['keys'][dba['options'][DBMapper.INDEX_PRIMARY]]
        primaryKeyB = dbb['keys'][dba['options'][DBMapper.INDEX_PRIMARY]]

        query = "select A.%s, B.%s from %s as A join %s as B on A.%s=B.%s" % (
                primaryKeyA, primaryKeyB, obja._table, objb._table, a, b)
        result = obja.__execute(cur, query, fetch=DBMapper.FETCH_ALL)

        load_pair_a = []
        load_pair_b = []
        for row in result:
            load_pair_a.append({primaryKeyA: row[0]})
            load_pair_b.append({primaryKeyB: row[1]})

        return (
            DBResultList([cls.load(**pair) for pair in load_pair_a]),
            DBResultList([cond.load(**pair) for pair in load_pair_b]),
        )

    def __init__(self, db, keys, keytypes, options=__default_options):

        self._db = db
        self._options = options
        if 'tableName' not in self._options:
            self._table = self.__class__.__name__.lower()
        else:
            self._options['tableName']

        self._keys = keys
        self._keytypes = keytypes

        self._primary_ind = self._options[DBMapper.INDEX_PRIMARY]
        self._autoincr_ind = self._options[DBMapper.INDEX_AUTOINCR]
        self._primary = self._keys[self._primary_ind]
        self._unique_keys = self._options[DBMapper.INDEX_UNIQUE]

        self.__generate_structure()
        self.__generate_getters()
        self.__generate_setters()
        self.__generate_properties()

    def __execute(self, cur, sql, fetch=FETCH_ONE, limit=-1, args=()):
        """ __execute(self,
                      cur      => pointer to database cursor
                      sql      => sql query to execute
                      fetch    => amount of results to fetch
                      limit    => query limit if not use FETCH_ONE
                      args     => query arguments to parse in)

            Filters, quotes, and executes the provided sql query and returns
            a list of database rows.
        """

        query = sql
        try:
            if len(args) >= 1:
                cur.execute("select " + ", ".join(["quote(?)" for i in args]),
                            args)
                quoted_values = cur.fetchone()
                for quoted_value in quoted_values:
                    query = query.replace('?', str(quoted_value), 1)
        except:
            pass

        try:
            cur.execute(query)
        except (sqlite3.ProgrammingError):
            try:
                cur.execute(query, args)
            except Exception as e:
                raise DBMapperException(
                    "Error while executing query [%s]" % (query), cause=e)
        except Exception as e:
            raise DBMapperException(
                "Error while executing query [%s]" % (query), cause=e)

        if fetch == DBMapper.FETCH_ONE:
            return cur.fetchone()
        elif fetch == DBMapper.FETCH_MANY:
            if limit == -1:
                limit = cur.arraysize
            return cur.fetchmany(size=limit)
        elif fetch == DBMapper.FETCH_ALL:
            return cur.fetchall()
        else:
            return cur.fetchall()

    def __get_table_info(self, table=None):
        """ __get_table_info(self, table)

            Returns pragma information for a table.
        """

        table = self._table if table is None else table
        cur = self._db.cursor()
        query = "pragma table_info(%s)" % (table)

        return self.__execute(cur, query, fetch=DBMapper.FETCH_ALL)

    def __generate_structure(self):
        """ __generate_structure(self)

            Generates table structure for determining column updates and
            search information.
        """

        # use pragma constructs to get table into
        tblinfo = self.__get_table_info()

        # create the table if the statement does not exist
        if len(tblinfo) == 0:
            ins = zip(self._keys, self._keytypes)
            typarr = []
            for pair in ins:
                if pair[0] == self._primary:
                    # identifier type primary key
                    if self._autoincr_ind:
                        typarr.append("%s %s primary key autoincrement" % (
                            pair[0], pair[1]))
                    else:
                        typarr.append("%s %s primary key" % (pair[0], pair[1]))
                elif pair[0] in self._unique_keys:
                    typarr.append("%s %s unique" % (pair[0], pair[1]))
                else:
                    # identifier type
                    typarr.append("%s %s" % (pair[0], pair[1]))
            cur = self._db.cursor()
            # create table if not exists <table> (<typarr>)
            query = "create table if not exists %s (%s)" % \
                (self._table, ', '.join(typarr))
            self.__execute(cur, query)

        # make sure table columns are up to date.
        if len(tblinfo) > 0:
            # use pragma table info to build database schema
            schema_ids = []
            schema_types = []
            for col in tblinfo:
                schema_ids.append(col[1])
                schema_types.append(col[2])
            # use schema to determine / apply database updates
            schema_updates = []
            for pair in zip(self._keys, self._keytypes):
                if pair[0] in schema_ids:
                    continue
                else:
                    schema_updates.append("%s %s" % (pair[0], pair[1]))
            for defn in schema_updates:
                query = "alter table %s add column %s" % (self._table, defn)
                cur = self._db.cursor()
                self.__execute(cur, query)

        # generate full text search table that corresponds with this dbo
        if self._options[DBMapper.GENERATE_FTS_VT]:
            if len(self.__get_table_info("_search_%s" % (self._table))) == 0:
                cur = self._db.cursor()
                # fts4 table doesn't exist, make it.
                query = "create virtual table _search_%s using fts4(%s, content='%s')" % \
                        (self._table, ','.join(self._keys), self._table)
                self.__execute(cur, query)
                # create pre/post update/delete triggers for cascading updates
                # XXX - [trigger warning] DO WE NEED THE TRIGGERS
                query = "create trigger _%s_bu before update on %s begin delete from _search_%s where docid=old.rowid; end;" % \
                        (self._table, self._table, self._table)
                self.__execute(cur, query)
                query = "create trigger _%s_bd before delete on %s begin delete from _search_%s where docid=old.rowid; end;" % \
                        (self._table, self._table, self._table)
                self.__execute(cur, query)
                search_keys = ','.join(['docid'] + self._keys[1:])
                target_keys = ','.join(['new.' + vkey for vkey in self._keys])
                query = "create trigger _%s_au after update on %s begin insert into _search_%s(%s) values(%s); end;" % \
                        (self._table, self._table, self._table, search_keys,
                         target_keys)
                self.__execute(cur, query)
                query = "create trigger _%s_ai after insert on %s begin insert into _search_%s(%s) values(%s); end;" % \
                        (self._table, self._table, self._table, search_keys,
                         target_keys)
                self.__execute(cur, query)

        self._db.commit()

    def __generate_getters(self):
        """ __generate_getters(self)

            Generates magical getter methods for pull data from the
            underlying database.
        """

        for _key in self._keys:
            def getter_templ(self, __key=_key):
                if __key not in self._keys:
                    return
                cur = self._db.cursor()
                # select * from table where key=<key>
                query = "select %s from %s where %s=?" % (
                    __key, self._table, self._primary)
                result = self.__execute(
                    cur,
                    query,
                    args=(getattr(self, "_%s" % (self._primary)),))
                try:
                    return result[0]
                except:
                    return result
            setattr(self, "get_" + _key, types.MethodType(getter_templ, self))

    def __generate_setters(self):

        for _key in self._keys:
            def setter_templ(self, value, __key=_key):
                if __key not in self._keys:
                    return
                cur = self._db.cursor()
                # update table set key=value where primary=id
                query = "update %s set %s=? where %s=?" % (
                    self._table, __key, self._primary)
                self.__execute(
                    cur,
                    query,
                    args=(value, getattr(self, "_%s" % (self._primary)),))
                self._db.commit()
                setattr(self, "_%s" % (__key), value)
            setattr(self, "set_" + _key, types.MethodType(setter_templ, self))

    def __generate_properties(self):

        for _key in self._keys:
            setattr(self, "_%s" % (_key), None)

        for _key in self._keys:
            getf = getattr(self, "get_%s" % (_key))
            setf = getattr(self, "set_%s" % (_key))
            setattr(self, _key, property(getf, setf, None, "[%s] property"))

    def create(self):

        cur = self._db.cursor()
        vals = []
        for key in self._keys:
            if key == self._primary and self._autoincr_ind:
                vals.append(None)  # Put None in for the index because autoinc
            else:
                vals.append(getattr(self, "_%s" % (key)))
        qst = ', '.join(["?" for item in vals])
        query = "insert into %s values (%s)" % (self._table, qst)
        self.__execute(cur, query, args=vals)
        setattr(self, "_%s" % (self._primary), cur.lastrowid)

    def delete(self):

        cur = self._db.cursor()

        qst = "%s=?" % (self._keys[self._primary_ind])
        primary_val = getattr(self, "_%s" % (self._keys[self._primary_ind]))

        query = "delete from %s where (%s)" % (self._table, qst)
        self.__execute(cur, query, args=(primary_val,))


class DBResultList(list):

    def __init__(self, extend=None):

        if isinstance(extend, list):
            for item in extend:
                if isinstance(item, DBMapper):
                    self.append(item)
                else:
                    continue

    def filter_equals(self, key, val):
        """ filter_equals(key, val) ->
              filters database find result based on
              key-value equality.
        """

        res = DBResultList()

        for dbo in self:
            try:
                if getattr(dbo, "_%s" % (key)) == val:
                    res.append(dbo)
                else:
                    continue
            except:
                continue

        return res

    def filter_iequals(self, key, val):
        """ filter_iequals(key, val) ->
              filters database find result based on
              case insensitive key-value equality.
              assumes that db attribute and val are strings.
        """

        res = DBResultList()

        for dbo in self:
            try:
                if getattr(dbo, "_%s" % (key)).lower() == val.lower():
                    res.append(dbo)
                else:
                    continue
            except:
                continue

        return res

    def filter_inequals(self, key, val):
        """ filter_inequals(key, val) ->
              filters database find result based on
              key-value inequality.
        """

        res = DBResultList()

        for dbo in self:
            try:
                if getattr(dbo, "_%s" % (key)) != val:
                    res.append(dbo)
                else:
                    continue
            except:
                continue

        return res

    def filter_regex(self, key, regex):
        """ filter_regex(key, regex) ->
              filters database find result based on
              regex value matching.
        """

        res = DBResultList()

        for dbo in self:
            try:
                if re.match(regex, getattr(dbo, "_%s" % (key))) is not None:
                    res.append(dbo)
                else:
                    continue
            except:
                continue

        return res


class DBMapperException(Exception):

    def __init__(self, message, cause=None):

        if cause is not None:
            super(DBMapperException, self).__init__(
                message + u', caused by ' + repr(cause))
        elif cause is None:
            super(DBMapperException, self).__init__(message)

        self.message = message
        self.cause = cause

    def __str__(self):

        return repr(self.message) + u', caused by ' + repr(self.cause)



# -*- coding: utf-8 -*-


class SharedState(object):
    """ This class is for meta-class use as a state machine for
        persistence so we don't use any singleton design.

        The module is "Borg-ish", as this implementation is loosely
        based on the Borg design pattern by Alex Martelli.
    """

    @classmethod
    def __initialize_states(cls):
        """ Initializes a class-scoped states dictionary to store named
            states.
        """

        if not hasattr(cls, "_SharedState__states"):
            cls.__states = {}

    def __init__(self, *args, **kw):
        """ Calls the classes state dict initializer and loads initial
            state, if provided.
        """

        self.__initialize_states()

        if "state" in kw:
            self.load_state(kw.get("state"))

    def load_state(self, state):
        """ Loads state into the class, overwriting all data that was
            previously stored.
        """

        if state in self.__states:
            self.__dict__ = self.__states[state]
        else:
            raise NameError("Can't load non-existent state '%s'." % (state))

    def save_state(self, state):
        """ Saves class state into a namespace on the class' shared state
            dict.
        """

        if state in self.__states:
            raise NameError("Can't overwrite stored state '%s'." % (state))
        self.__states.update({state: self.__dict__})



# -*- coding: utf-8 -*-
import json
import time
import types
import uuid
from difflib import SequenceMatcher

from malibu.util.decorators import function_marker

""" Brine is a play on Python's pickle module, which is used for
    serializing data. Brine is used for serialization as well, but
    into JSON, not a binary structure.
"""

# Declare a set of method types that should be filtered for.
METHOD_TYPES = [types.MethodType, types.FunctionType, types.LambdaType]
__brine_nested_method = function_marker("_brine", "nested")


def fuzzy_ratio(a, b):
    """ Compares two values using the SequenceMatcher from difflib.
        Used for ~approximated~ fuzzy search.
    """

    return SequenceMatcher(None, a, b).ratio()


def nested_object(target_cls):
    """ Used to do late deserialization of nested brine objects.
    """

    @__brine_nested_method
    def nested_dec(data):

        return target_cls.by_json(data)

    return nested_dec


class BrineObject(object):
    """ This object is for use as a base class for other data.
        Essentially, it will expose a set of members that can be set
        and then squashed down to a JSON object through a call to to_json.

        It can also be used as a meta-class for the base of a caching object
        model or other neat things.
    """

    @classmethod
    def by_json(cls, data, **kw):
        """ Creates a new instance and calls from_json on the instance.

            Will take kwargs and pass to the underlying instance
            initializer.
        """

        inst = cls(**kw)
        inst.from_json(data)

        return inst

    def __init__(self, *args, **kw):

        # Do this because MRO.
        super(BrineObject, self).__init__()

        # For now, lets make this simple and treat fields with no special
        # syntax (underlines, mainly) as our schema.
        self._special_fields = ["timestamp", "uuid"]
        self._fields = []
        for field in dir(self):
            if field.startswith("_"):
                continue
            # Also, make sure this isn't a function.
            if type(getattr(self, field)) in METHOD_TYPES:
                attrm = getattr(self, field)
                if getattr(attrm, "_brine", None) == "nested":
                    self._fields.append(field)
                continue
            self._fields.append(field)

        if kw.get("timestamp", False):
            self.timestamp = int(time.time())

        if kw.get("uuid", False):
            self.uuid = str(uuid.uuid4())

    def as_dict(self):
        """ Returns the dictionary representation of the fields
            in this object.
        """

        obj = {}

        for val in self._fields + self._special_fields:
            if not hasattr(self, val):
                continue
            # Also, make sure this isn't a function.
            if type(getattr(self, val)) in METHOD_TYPES:
                continue
            attr = getattr(self, val)
            if isinstance(attr, BrineObject):
                obj.update({val: attr.as_dict()})
            else:
                obj.update({val: getattr(self, val)})

        return obj

    def to_json(self):
        """ Converts the object into JSON form.
            Simple, right?
        """

        return json.dumps(self.as_dict())

    def from_json(self, data):
        """ Converts the JSON data back into an object, then loads
            the data into the model instance.
        """

        obj = json.loads(data)

        if not isinstance(obj, dict):
            raise TypeError("Expected JSON serialized dictionary, not %s" % (
                type(obj)))

        for k, v in obj.items():
            # We need to make sure the data is sanitized a little bit.
            if k.startswith("_") and k not in self._special_fields:
                continue
            if k in self._fields:
                fval = getattr(self, k, None)
                if getattr(fval, "_brine", None) == "nested":
                    # this is a nested object, deserialize the data
                    setattr(self, k, fval(v))
                else:
                    setattr(self, k, v)


class CachingBrineObject(BrineObject):
    """ This is a magical class that performs the same function as the
        BrineObject, but it also adds object caching, searching, and fuzzy
        searching on the cache. Also provided is cached field invalidation /
        "dirtying".
    """

    # Ratio for fuzzy search. Closer to 1.0 means stricter results.
    _FUZZ_RATIO = 0.535

    @classmethod
    def __initialize_cache(cls):
        """ Initialize a class-level cache to store Json models for cache and
            searching purposes.
        """

        if not hasattr(cls, "_CachingBrineObject__cache"):
            cls.__cache = []

    @classmethod
    def fuzzy_search(cls, ignore_case=False, **kw):
        """ Performs a fuzzy search on the cache to find objects that have at
            least a diff ratio of FUZZ_RATIO.

            Note that this can return more than one object and it may not be
            accurate. Time will tell.

            Returns a list of matches ordered by likelihood of match.
        """

        ratios = {}

        for k, v in kw.items():
            for obj in cls.__cache:
                ob_value = getattr(obj, k, None)
                if ignore_case:
                    if isinstance(v, str) and isinstance(ob_value, str):
                        r = fuzzy_ratio(ob_value.lower(), v.lower())
                else:
                    r = fuzzy_ratio(ob_value, v)
                if r >= cls._FUZZ_RATIO:
                    ratios.update({obj: r})

        # TODO - sort by fuzzy search ratio.
        # We need to ensure the results get properly sorted by match ratio
        # before returning.

        return ratios.keys()

    @classmethod
    def search(cls, ignore_case=False, **kw):
        """ Searches through the cache to find objects with field that match
            those given by the **kw.

            Note that this can return more than one object.
        """

        result = []

        for k, v in kw.items():
            for obj in cls.__cache:
                ob_value = getattr(obj, k, None)
                if ignore_case:
                    if isinstance(v, str) and isinstance(ob_value, str):
                        r = (v.lower() == ob_value.lower())
                else:
                    r = (v == ob_value)
                if r:
                    if obj in result:
                        continue
                    else:
                        result.append(obj)
                else:
                    continue

        return result

    def __init__(self, *args, **kw):

        # Call the parent initializer.
        super(CachingBrineObject, self).__init__(self, *args, **kw)
        self._initialized = False

        # Make sure the cache is initialized.
        self.__initialize_cache()

        # The "dirty" cache list is just a list of fields that have been
        # updated.
        self.__dirty = []

        # Throw this object into the cache.
        self.__cache.append(self)

        # Let the attribute handler know we're done loading.
        self._initialized = True

    def __setattr__(self, attr, value):
        """ Sets local fields and determines if the cache needs to be
            marked dirty for that set and ensures that the value can
            actually be set.
        """

        # I wish this didn't have to be a special case.
        if attr == "_initialized":
            self.__dict__[attr] = value
            return

        # Check that init has finished.
        if not getattr(self, "_initialized", False):
            self.__dict__[attr] = value
            return

        # Check various conditions used to determine if a variable has been
        # dirtied or can be set.
        if attr in self._fields:
            if attr not in self.__dirty:
                self.__dirty.append(attr)
        elif attr in self._special_fields:
            raise AttributeError("Special field {} is immutable.".format(attr))
        elif attr not in self.__dict__:
            raise AttributeError("Field {} does not exist.".format(attr))

        # Verify that the set *will not* overwrite a method.
        _attr_cur = getattr(self, attr)
        if type(_attr_cur) in METHOD_TYPES:
            raise TypeError("Function {} can not be overwritten.".format(attr))

        # Set the variable in the dictionary.
        self.__dict__[attr] = value

    def uncache(self):
        """ Removes the object from the state cache forcibly.
        """

        self.__cache.remove(self)

    def unmark(self, *fields):
        """ Unmarks some field as dirty. Should only be called after
            the upstream is updated or only if you know what you're doing!
        """

        for field in fields:
            if field not in self.__dirty:
                continue

            self.__dirty.remove(field)

    def dirty_dict(self):
        """ Dumps a dictionary of dirty fields.
        """

        obj = {}
        for val in self.__dirty:
            if not hasattr(self, val):
                continue
            # Also, make sure this isn't a function.
            if type(getattr(self, val)) in METHOD_TYPES:
                continue
            obj.update({val: getattr(self, val)})

        return obj

    def dirty_json(self):
        """ Dumps the dirty dictionary as JSON.
        """

        return json.dumps(self.dirty_dict())



# -*- coding: utf-8 -*-
import io


class TextTable(object):

    TABLE_CORNER = "+"
    TABLE_VT_BORDER = "|"
    TABLE_HZ_BORDER = "-"

    def __init__(self, min_width=12):

        self.min_width = min_width

        self._rows = 0
        self._columns = 0

        # Header data should be simple; it should be no more than a list of
        # strings naming each column header.
        self._header_data = []

        # Row data should be internally stored in a zipped-set form, or, as a
        # list of tuples, which each tuple containing all the entries for a
        # single row.
        self._row_data = []

    def add_header_list(self, el):

        if not isinstance(el, list):
            return

        self._columns = len(el)
        self._header_data = el

    def add_header(self, *args):

        self._columns = len(args)
        self._header_data = [arg for arg in args]

    def add_data_dict(self, el):

        """ add_data_dict only really makes sense to use when there is a
            single pair mapping (eg., key, value) or a two column
            display.  If that is not that case, add_data_ztup is a better
            choice.
        """

        if not isinstance(el, dict):
            return

        self._rows = len(el)

        for key, value in el.items():
            self._row_data.append((key, value,))

    def add_data_ztup(self, el):

        """ add_data_ztup will take any as much data as you need and can even
            fill the place of add_data_dict.  add_data_ztup takes a list of
            tuples. each tuple should contain a row of elements, one element
            for each column. essentially, the argument that will be passed
            in should look like:
              [
                (x1, y1, z1),
                (x2, y2, z2),
                    ...
                (xn, yn, zn)
              ]
        """

        if not isinstance(el, list):
            return

        self._rows = len(el)

        for row in el:
            row = [str(item) for item in row]
            if len(row) > self._columns:
                row = row[0:self._columns - 1]
            elif len(row) < self._columns:
                col_diff = (self._columns - len(row)) + 1
                col_fill = len(row)
                for idx in range(col_fill, col_diff):
                    el[idx] = ''
            self._row_data.append(row)

    def add_data_kv(self, k, v):

        """ add_data_kv is a simplified add_data_dict for pushing a data pair onto the
            row list on-the-fly. only effective for two-column data sets.
        """

        self._row_data.append((k, v,))

    def add_data_list(self, el):

        """ add_data_list adds a list of data to the table. this is primarily
            suitable for single-column data sets.
        """

        [self._row_data.append((elm,)) for elm in el]

    def add_data_csv_file(self, fobj):

        """ add_data_csv_file loads data from a comma-separated value file.
            the first row is the header, everything else is actual data.
            works if the file is provided or if a string containing the
            filename is given.
        """

        if isinstance(fobj, str):
            try:
                fobj = io.open(fobj, 'r')
            except:
                raise

        data_flag = False

        for line in fobj.readlines():
            line = line.strip().split(",")
            if not data_flag:
                self._columns = len(line)
                self.add_header(*line)
                data_flag = True
                continue
            self._row_data.append(tuple(line))

    def __transpose_list(self, li):

        """ performs a simple transposition on a list.
            used for calculating row sizes and maxes.
        """

        return map(list, zip(*li))

    def __calculate_row_max(self):

        """ calculates the max size of the rows so the table has uniform column sizes.
        """

        __sizes = []

        for row in self._row_data:
            __sizes.append([len(el) for el in row])

        __sizes = self.__transpose_list(__sizes)

        return [max(cols) for cols in __sizes]

    def __format_divider(self):

        """ returns the table divider.
        """

        s = ""
        col_sizes = self.__calculate_row_max()

        for size in col_sizes:
            s += TextTable.TABLE_CORNER
            if size >= self.min_width:
                s += (TextTable.TABLE_HZ_BORDER * (size + 2))
            else:
                s += (TextTable.TABLE_HZ_BORDER * self.min_width)
        s += TextTable.TABLE_CORNER

        return s

    def __pad_left(self, txt, length):

        """ pads a string to be length characters long, from left.
        """

        s = []

        if len(txt) < length:
            diff = length - len(txt) - 1
            [s.append(" ") for i in range(0, diff)]
            s.append(''.join(txt) + " ")
        elif len(txt) == length:
            s.append(txt)

        return ''.join(s)

    def __format_header_data(self):

        """ formats the header data.
        """

        lines = []
        col_sizes = self.__calculate_row_max()

        line = ""

        cd = zip(self._header_data, col_sizes)
        for (text, size) in cd:
            if size < self.min_width:
                size = self.min_width
            elif size >= self.min_width:
                size = size + 2
            text = self.__pad_left(text, size)
            line += TextTable.TABLE_VT_BORDER
            line += text
        line += TextTable.TABLE_VT_BORDER
        lines.append(line)

        return lines

    def __format_table_data(self):

        """ formats the table data.
        """

        lines = []
        col_sizes = self.__calculate_row_max()

        for row in self._row_data:
            rd = zip(row, col_sizes)
            line = ""
            for (text, size) in rd:
                if size < self.min_width:
                    size = self.min_width
                elif size >= self.min_width:
                    size = size + 2
                text = self.__pad_left(text, size)
                line += TextTable.TABLE_VT_BORDER
                line += text
            line += TextTable.TABLE_VT_BORDER
            lines.append(line)

        return lines

    def format(self):

        """ format the table and return the string.
        """

        lines = []
        divider = self.__format_divider()

        lines.append(divider)
        [lines.append(line) for line in self.__format_header_data()]
        lines.append(divider)
        [lines.append(line) for line in self.__format_table_data()]
        lines.append(divider)

        return lines



# -*- coding: utf-8 -*-
from __future__ import print_function

import glob
import os


modules = glob.glob(os.path.dirname(__file__) + "/*.py")
__all__ = [os.path.basename(f)[:-3] for f in modules
           if not os.path.basename(f).startswith('_') and
           not f.endswith('__init__.py') and os.path.isfile(f)]


def unicode_type():
    """ Find the Unicode class for this version of Python.
        Between Python 2.x and 3.x, classes and text handling has
        changed a considerable amount.
        Python 2.x uses ASCII strings with str() and Unicode with
        unicode().
        Python 3.x uses Unicode strings with str() and "strings of
        bytes" as bytes().
    """

    try:
        import builtins
        __old_unicode = False
    except ImportError:
        __old_unicode = True

    if __old_unicode:
        return unicode  # flake8: noqa
    elif not __old_unicode:
        return str  # flake8: noqa


def string_type():
    """ Find the String class for this version of Python.
        Between Python 2.x and 3.x, classes and text handling has
        changed a considerable amount.
        Python 2.x uses ASCII strings with str() and Unicode with
        unicode().
        Python 3.x uses Unicode strings with str() and "strings of
        bytes" as bytes().
    """

    try:
        import builtins
        __old_string = True
    except ImportError:
        __old_string = False

    if __old_string:
        return str  # flake8: noqa
    elif not __old_string:
        return bytes  # flake8: noqa


def unicode2str(obj):
    """ Recursively convert an object and members to str objects
        instead of unicode objects, if possible.

        This only exists because of the incoming world of unicode_literals.

        :param object obj: object to recurse
        :return: object with converted values
        :rtype: object
    """

    if isinstance(obj, dict):
        return {unicode2str(k): unicode2str(v) for k, v in
                obj.items()}
    elif isinstance(obj, list):
        return [unicode2str(i) for i in obj]
    elif isinstance(obj, unicode_type()):
        return obj.encode("utf-8")
    else:
        return obj


def str2unicode(obj, encoding='utf-8'):
    """ Recursively convert an object and members to unicode objects
        instead of str objects, if possible.

        This only exists because of the incoming world of unicode_literals.

        :param object obj: object to recurse
        :param str encoding: encoding to decode from
        :return: object with converted values
        :rtype: object
    """

    if isinstance(obj, dict):
        return {str2unicode(k): str2unicode(v) for k, v in
                obj.items()}
    elif isinstance(obj, list):
        return [str2unicode(i) for i in obj]
    elif isinstance(obj, string_type()):
        return obj.decode(encoding)
    else:
        return obj



# -*- coding: utf-8 -*-
from __future__ import print_function

import types

__doc__ = """
    This module contains decorator generators.
    Essentially this will be a medley of delicious functions
    to generate relatively useful, reusable, generic decorators for code.
"""


def function_registrator(target):
    """ function_registrator generates a simple decorator that will
        take a function with any set of arguments and register that
        function within a target list.
    """

    def decorator(func):
        """ This is a "flexible" decorator function that pushes to
            target thanks to scope magic.
        """

        if func not in target:
            target.append(func)

        return func

    return decorator


def function_marker(attr, value):
    """ function_marker generates a simple decorator that will
        take a function with any set of arguments and set a given
        attribute on that function with setattr().
    """

    def decorator(func):
        """ This is a "flexible" decorator function that sets the
            attribute on the target function.
        """

        setattr(func, attr, value)

        return func

    return decorator


def function_kw_reg(target, req_args):
    """ function_kw_reg generates a more complex decorator that
        must be used with the argument names given in req_args.
        useful for attribute-based assertion.

        NOTE: target should be a dict-typed class

        example:

            trait = function_kw_reg(target, ['val1', 'val2', 'val3'])

            @trait(val1 = "attr1", val2 = "attr2", val3 = "attr3")
            def do_thing(...):
                pass
    """

    def decorator_outer(**kw):

        for req in req_args:
            if req not in kw.keys():
                raise KeyError("Missing required attribute: %s" % (req))

        def decorator_inner(func):

            if func not in target:
                target.update({func: kw})

            return func

        return decorator_inner

    return decorator_outer


def function_intercept_scope(*fs, **kwa):
    """ function_intercept_scope allows the injection of variables
        into a decorated function from a specified function or a
        list of functions. Variables injected are based on the return
        value of the function(s) provided.

        Provided functions should return a dictionary of kv pairs, with
        the key being a string and the value being any object to be injected.

        If intercept_args is True, *args and **kw going to the function
        be sent to the intercepting functions first.

        example:

            def intercept_argparser_inst(*args, **kw):

                if 'args' in kw:
                    argparser = kw['args']
                else:
                    argparser = CommandModuleLoader(state="default") \
                                .get_argument_parser()

                return {
                    "argparser": argparser
                }

            command_func = function_intercept_scope(
                intercept_argparser_inst,
                intercept_args=True)

            class ExampleModule(module.CommandModule):
                # ... all command module init, etc ...

                @command_func
                def command_do(self, *args, **kw):
                    # 'argparser' will be injected here as a global
                    # do stuff ...
    """

    intercept_args = False

    if 'intercept_args' in kwa:
        intercept_args = kwa['intercept_args']

    _funcs = []
    for fun in fs:
        if type(fun) in [types.FunctionType, types.LambdaType]:
            _funcs.append(fun)

    def decorator_outer(func):

        def decorator_inner(*args, **kw):

            fg = func.func_globals
            restore = {}
            _sentinel = object()

            inject = {}
            for fun in _funcs:
                try:
                    if intercept_args:
                        fi = fun(*args, **kw)
                    else:
                        fi = fun()
                    if not isinstance(fi, dict):
                        continue

                    inject.update(fi)
                except Exception as e:
                    print("Encountered an Exception while running interceptor:", fun)
                    print("Exception:", e)
                    continue

            for var in inject.keys():
                restore[var] = inject.get(var, _sentinel)
                fg[var] = inject[var]

            try:
                res = func(*args, **kw)
            finally:
                for var in inject.keys():
                    if restore[var] is _sentinel:
                        del fg[var]
                    else:
                        fg[var] = restore[var]

            return res

        return decorator_inner

    return decorator_outer



# -*- coding: utf-8 -*-
from malibu import command
from malibu import config
from malibu import database
from malibu import design
from malibu import text
from malibu import util


__version__ = '0.1.6'
__doc__ = """
malibu is a collection of classes and utilities that make writing code
a little bit easier and a little less tedious.

The whole point of this library is to have a small codebase that could
be easily reused across projects with nice, easily loadable chunks that
can be used disjointly.
"""



# -*- coding: utf-8 -*-
import glob
import inspect
import os

modules = glob.glob(os.path.dirname(__file__) + "/*.py")
__all__ = [os.path.basename(f)[:-3] for f in modules
           if not os.path.basename(f).startswith('_') and not
           f.endswith('__init__.py') and os.path.isfile(f)]


def get_caller():
    """ Inspects the call stack to determine the path of a caller.
        Returns a string representing the full path to the caller.
    """

    # If we want to get the "caller" of a function, from this context we
    # must go backwards by two layers. The first frame is the current frame.
    # The second frame would be the function that calls get_caller().
    # The third frame is the frame we want to know the identity of.

    # The named tuple, Traceback, looks like this:
    #  Traceback(filename, lineno, function, code_context, index)

    frame = inspect.currentframe()
    # Get a list of frames up to two outside of the current frame1.
    callstack = inspect.getouterframes(frame, 2)
    # callstack is [Traceback(...), ...]
    # Grab the outermost frame
    caller = callstack[2][0]
    # Pull the Traceback tuple for the current frame.2
    callerinfo = inspect.getframeinfo(caller)

    # Determine if the function is a bound method.
    if 'self' in caller.f_locals:
        # Since the caller is a bound method, extract the 'self' reference
        # from the locals, then pull the class name from the instance class.
        caller_class = caller.f_locals['self'].__class__.__name__
    else:
        # There is no object this method is bound to -- it is a module
        # function.
        caller_class = None

    # Get the FQN of the module that the calling frame belongs to.
    caller_module = inspect.getmodule(caller).__name__
    # callerinfo is Traceback(...)
    caller_name = callerinfo[2]

    # Check the determined caller class.
    if caller_class:
        # If there is a caller class, prepend it to the caller name.
        caller_string = "%s.%s" % (caller_class, caller_name)
    else:
        # The caller string is just the caller name by itself.
        caller_string = "%s" % (caller_name)

    # Check the determined caller module.
    if caller_module:
        # Prepend the caller module to the caller string.
        caller_string = "%s." % (caller_module) + caller_string

    return caller_string


def get_calling_frame():

    frame = inspect.currentframe()
    callstack = inspect.getouterframes(frame, 2)
    caller = callstack[2][0]

    return caller


def get_current():

    frame = inspect.currentframe()
    callstack = inspect.getouterframes(frame, 1)
    frame = callstack[1][0]
    frameinfo = inspect.getframeinfo(frame)

    if 'self' in frame.f_locals:
        current_class = frame.f_locals['self'].__class__.__name__
    else:
        current_class = None

    current_module = inspect.getmodule(frame).__name__
    current_name = frameinfo[2]

    if current_class:
        current_string = "%s.%s" % (current_class, current_name)
    else:
        current_string = "%s" % (current_name)

    if current_module:
        current_string = "%s." % (current_module) + current_string

    return current_string


def get_current_frame():

    frame = inspect.currentframe()
    callstack = inspect.getouterframes(frame, 1)
    caller = callstack[1][0]

    return caller



# -*- coding: utf-8 -*-
import glob
import os


modules = glob.glob(os.path.dirname(__file__) + "/*.py")
__all__ = [os.path.basename(f)[:-3] for f in modules
           if not os.path.basename(f).startswith('_') and
           not f.endswith('__init__.py') and os.path.isfile(f)]

__doc__ = """
malibu's database classes were mainly an experiment with ORM tech using
    Python's introspection capabilities, closures, and properties.

The actual ORM class exists as malibu.database.dbmapper.DBMapper
    and should be inherited to be used properly.

malibu.database.dbmapper.dbtypeconv is a stub module for installing
    adapters into the sqlite3 module.

As of the 0.1.6 release, the DBMapper and dbtypeconv are both
    deprecated in favour of external ORM projects with better compatibility.

.. automodule:: malibu.database.dbmapper
   :members:

.. automodule:: malibu.database.dbtypeconv
   :members:

"""



# -*- coding: utf-8 -*-
import random


__adjectives__ = [
    "smelly", "firey", "naughty", "sad", "frazzled", "excited",
    "gross", "chipper", "nasty", "nosey", "pokey", "frustrated", "glum",
    "bouncy", "electric", "immune", "excited", "perfunctory", "likable",
    "fastidious", "adorable", "charming", "pleasant", "crass", "gloomy",
    "faithful", "spectacular", "diminished", "hurtful", "voltaic",
    "fervent", "ravishing", "riveting", "peculiar", "cross", "smashed",
    "ridiculous", "pontiferous", "offended", "overzealous", "jealous",
    "freaky", "comical", "creepy", "disturbed", "massive", "bigoted",
    "ecstatic", "rowdy", "friendly", "enlightened", "failing"
]

__nouns__ = [
    "feynman", "einstein", "roadrunner", "tacocat", "grump", "fish",
    "michael", "bishop", "charlie", "metcalfe", "babbage", "zuckerberg",
    "nicholas", "priest", "doge", "pimp", "drone", "worker", "voltorb",
    "bulbasaur", "king", "bay", "transformers", "johnson", "baby",
    "charmander", "mew", "axel", "wayne", "charles", "jessica", "clarice",
    "doctor", "horton", "ballmer", "jobs", "gates", "failfish"
]


def get_simple_name(delim='_'):

    adj = random.choice(__adjectives__)
    noun = random.choice(__nouns__)

    return delim.join([adj, noun])



# -*- coding: utf-8 -*-
import abc
from datetime import datetime, timedelta

from malibu.config import configuration
from malibu.design import borgish
from malibu.util.decorators import function_registrator

__JOB_STORES__ = []
job_store = function_registrator(__JOB_STORES__)


class Scheduler(borgish.SharedState):

    def __init__(self, store='volatile', *args, **kw):

        super(Scheduler, self).__init__(*args, **kw)

        # Make sure the job store doesn't get reinitialized after loading
        # state through the SharedState mixin.
        if "state" not in kw:
            job_store = list(filter(lambda st: st.TYPE == store, __JOB_STORES__))
            if not job_store or len(job_store) == 0:
                raise SchedulerException(
                    "Could not find a job store for type: %s" % (store))
            elif len(job_store) > 1:
                raise SchedulerException(
                    "Selected more than one job store for type: %s" % (store))

            job_store = job_store[0]
            self._job_store = job_store(self)

    @property
    def job_store(self):

        return self._job_store

    def create_job(self, name, func, delta, recurring=False):
        """ Creates a new job instance and attaches it to the scheduler.

            Params
            ------
            name : str
                Name of the job to create.
            func : function
                Function that is specified as the execution callback.
            delta : datetime.timedelta
                Timedelta to execute on.
            recurring : bool, optional
                Whether the job should continue running after first run.

            Raises
            ------
            SchedulerException
                If the job already exists, func is none, or delta is not
                an instance of datetime.timedelta.

            Returns
            -------
            SchedulerJob
                If job creation was succesful.
        """

        if self.job_store.get_job(name):
            raise SchedulerException("Job already exists; remove it first.")
        if func is None:
            raise SchedulerException("Callback function is non-existent.")
        if not isinstance(delta, timedelta):
            raise SchedulerException(
                "Argument 'delta' was not a timedelta instance.")

        job = SchedulerJob(name, func, delta, recurring)
        self.add_job(job)

        return job

    def add_job(self, job):
        """ Adds a job to the list of jobs maintained by the scheduler.

            Params
            ------
            job : SchedulerJob
                Job to add to the list of active jobs.

            Raises
            ------
            SchedulerException
                If the job already exists in the jobs dictionary.
        """

        if self.job_store.get_job(job.get_name()):
            raise SchedulerException("Job already exists; remove it first.")

        job.begin_ticking()
        self.job_store.store(job)

    def remove_job(self, name):
        """ Removes a job from the list of jobs maintained by the scheduler.

            Params
            ------
            name : str
                Name of the job to remove from the job list.

            Raises
            ------
            SchedulerException
                If the job does not exist.
        """

        if not self.job_store.get_job(name):
            raise SchedulerException("Job does not exist.")

        self.job_store.destore(self.job_store.get_job(name))

    def tick(self):
        """ Gets the current time and checks the ETA on each job.
            If the job is ready, it executes and captures any exception
            that should raise out of the execute call.
            If an exception is captured, the job's onfail callbacks will be
            fired with a reference to the job object that was being triggered.
        """

        now = datetime.now()

        mark_removal = []

        for job in self.job_store.get_jobs():
            if job.is_ready(now):
                try:
                    job.execute()
                    job.set_traceback(None)
                except Exception as e:
                    job.set_traceback(e)
                    job.fire_onfail()

                mark_removal.append(job)

        for job in mark_removal:
            self.remove_job(job.get_name())
            if job.is_recurring():
                self.add_job(job)


class SchedulerJobStore(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, scheduler, *args, **kw):
        """ Initializes the job store.
        """

        self._scheduler = scheduler

        if "config" in kw:
            self.initialize(kw.get("config"))

    @abc.abstractmethod
    def initialize(self, config):
        """ Provides initialization from a Configuration object for job store
            implementations that need it.
        """

        if not isinstance(config, configuration.ConfigurationSection):
            raise TypeError("Expected instance of malibu.config."
                            "Configuration.ConfigurationSection, not %s" % (
                                config.__class__.__name__))

        return

    @staticmethod
    def updates_store(func):
        """ Decorator that forces an update of a job in the store after the
            decorated function is run.
        """

        def _funcarg_decorator(*args, **kw):
            """ This function chain will call the given function and then
                attempt to re-store the job in the job store.
            """

            func(*args, **kw)

            # Since we don't know the location of the job in the function
            # arguments, we can look in two places to sort-of enforce a
            # structure:
            #  - In the first *args slot
            #  - In the 'job' **kw slot.
            job = None
            if len(args) > 0:
                job = args[0]

            if not job or not isinstance(job, SchedulerJob):
                job = kw.get('job', None)

            if not job or not isinstance(job, SchedulerJob):
                # Best we can do here is just return or warn.
                return

            scheduler = job._scheduler
            store = scheduler.job_store
            store.store(job, update=True)

        return _funcarg_decorator

    @abc.abstractmethod
    def get_jobs(self):
        """ Returns all jobs in the job store.
        """

        return

    @abc.abstractmethod
    def get_job(self, job_name):
        """ Returns a job from the store with the given name.
        """

        return

    @abc.abstractmethod
    def store(self, job, update=False):
        """ Serializes a job into the job store.
        """

        if not isinstance(job, SchedulerJob):
            raise TypeError("Job argument is not an instance of SchedulerJob")

        return

    @abc.abstractmethod
    def destore(self, job):
        """ Removes a job from the job store.
        """

        if not isinstance(job, SchedulerJob):
            raise TypeError("Job argument is not an instance of SchedulerJob")

        return


@job_store
class VolatileSchedulerJobStore(SchedulerJobStore):
    """ Implements a purely in-memory job store backed by a dictionary.

        This job store type does not need a set of configurations pass in to
        it. Any jobs pushed into this store will be lost when the application
        is shut down!
    """

    TYPE = 'volatile'

    def __init__(self, scheduler):

        super(VolatileSchedulerJobStore, self).__init__(scheduler)

        self.__jobs = {}

    def initialize(self, config):

        return

    def get_jobs(self):

        return self.__jobs.values()

    def get_job(self, job_name):

        return self.__jobs.get(job_name, None)

    def store(self, job, update=False):

        super(VolatileSchedulerJobStore, self).store(job, update)

        if job.get_name() in self.__jobs and not update:
            return False

        self.__jobs.update({job.get_name(): job})

        return True

    def destore(self, job):

        super(VolatileSchedulerJobStore, self).destore(job)

        if job.get_name() not in self.__jobs:
            return False

        return True if self.__jobs.pop(job.get_name()) else False


class SchedulerJob(object):

    def __init__(self, name, function, delta, state, recurring=False):

        self._scheduler = Scheduler()
        self._name = name
        self._function = function
        self._delta = delta
        self._recurring = recurring
        self._last_traceback = None
        self._onfail = []
        self._metadata = {}
        self.onfail = function_registrator(self._onfail)

        self._eta = delta

    @property
    def metadata(self):

        return self._metadata

    def get_name(self):

        return self._name

    def get_eta(self):

        return self._eta

    def get_traceback(self):

        return self._last_traceback

    @SchedulerJobStore.updates_store
    def set_traceback(self, stack):

        self._last_traceback = stack

    @SchedulerJobStore.updates_store
    def attach_onfail(self, func):

        self._onfail.append(func)

    @SchedulerJobStore.updates_store
    def detach_onfail(self, func):

        self._onfail.remove(func)

    def fire_onfail(self):

        for callback in self._onfail:
            callback(job=self)

    def is_recurring(self):

        return self._recurring

    def is_ready(self, time):

        if time >= self._eta:
            return True
        else:
            return False

    @SchedulerJobStore.updates_store
    def begin_ticking(self):

        self._eta = datetime.now() + self._delta

    @SchedulerJobStore.updates_store
    def execute(self):

        self._function()

        if self._recurring:
            self._eta += self._delta


class SchedulerException(Exception):

    def __init__(self, value):

        self.value = value

    def __str__(self):

        return repr(self.value)



# -*- coding: utf-8 -*-
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup

import malibu

setup(
    name='malibu',
    version=malibu.__version__,
    description="maiome library of utilities",

    url="https://phabricator.ramcloud.io/tag/malibu",
    author="Sean Johnson",
    author_email="sean.johnson@maio.me",

    license="Unlicense",

    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: Public Domain",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    packages=['malibu',
              'malibu.command',
              'malibu.config',
              'malibu.database',
              'malibu.design',
              'malibu.text',
              'malibu.util'],
    package_dir={'malibu': 'malibu'},
    install_requires=[
        'dill'
    ],
    test_suite='nose.collector',
    tests_require=[
        'nose',
        'coverage'
    ],
    zip_safe=True
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
import io

from contextlib import closing
try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen

from malibu.text import unicode_type, unicode2str


__doc__ = """
malibu.config.configuration
---------------------------

INI-style configuration implementation with some special features
to make configuration a little simpler.

"""


class ConfigurationSection(dict):
    """ The ConfigurationSection class is a modified dictionary that
        provides "helpers" to grab a configuration entry in it's correct
        "type" form.
    """

    def __init__(self):

        dict.__init__(self)

        self.mutable = True

    def __getitem__(self, key):

        try:
            return dict.__getitem__(self, key)
        except (IndexError, KeyError):
            raise KeyError("Unknown configuration key '%s'." % (key))

    def __setitem__(self, key, value):

        if self.mutable:
            self.update({key: value})
        else:
            raise AttributeError("This section is not mutable.")

    def set_mutable(self, mutable):
        """ Enforces immutability on a configuration section.
        """

        self.mutable = mutable

    def set(self, key, value):
        """ Allows programmatic setting of configuration entries.
        """

        return self.__setitem__(key, value)

    def get(self, key, default=None):
        """ The bare "get" on the underlying dictionary that returns the
            configuration entry in whatever form it was parsed as, typically
            a string.
        """

        try:
            return self.__getitem__(key)
        except IndexError:
            return default

    def get_list(self, key, delimiter=",", strip=True, default=[]):
        """ Attempts to take a something-delimited string and "listify" it.
            If an error occurs while attempting to listify, :param default:
            will be returned.
        """

        try:
            val = self.get(key)
            if isinstance(val, list):
                return val
            l = val.split(delimiter) if len(val) > 0 else default
            if strip:
                return [item.strip() for item in l]
            else:
                return l
        except:
            return default

    def get_string(self, key, default=""):
        """ Attempts to take the value stored and retrieve it safely as a
            string. If the value mapped to by :param key: is "!None", the
            object returned is NoneType.

            If an error occurs while trying to safely retrieve the string,
            :param default: is returned.
        """

        try:
            if str(self.get(key)) == '!None':
                return None
            return str(self.get(key)) or default
        except:
            return default

    def get_int(self, key, default=None):
        """ Attempts to fetch and intify the value mapped to by
            :param key:. If an error occurs while trying to intify the value,
            :param default: will be returned.
        """

        try:
            return int(self.get(key)) or default
        except:
            return default

    def get_bool(self, key, default=False):
        """ Attempts to safely fetch the value mapped to by :param key:.
            After successful retrieval, a conditional coercion to boolean
            is attempt. If the coercion to boolean fails, :param default: is
            returned.
        """

        try:
            val = self.get(key) or default
            if isinstance(val, bool):
                return val
            elif isinstance(val, unicode_type()):
                if val.lower() == 'true':
                    self.set(key, True)
                    return True
                elif val.lower() == 'false':
                    self.set(key, False)
                    return False
                else:
                    return default
            elif isinstance(val, int):
                if val == 1:
                    self.set(key, True)
                    return True
                elif val == 0:
                    self.set(key, False)
                    return False
                else:
                    return default
            else:
                print("get_bool: unknown value type")
                print("type(val) => ", type(val))
                print("unicode_type() => ", unicode_type())
                return default
        except:
            return default


class SectionPromise(object):
    """ this is a configuration section promise
        to make resolution of linked sections post-load
        easier.
    """

    promises = []

    def __init__(self, config, section, key, link):

        self.config = config
        self.section = section
        self.key = key
        self.link = link
        self.__fulfilled = False

        SectionPromise.promises.append(self)

    def __str__(self):
        """ Convert directly to a string for recreating the link
            during config write.  Better for serialization.
        """

        return '@' + self.link

    def resolve(self):
        """ Resolves a SectionPromise into the proper dictionary
            value.
        """

        if self.__fulfilled:
            return

        section = self.config.get_section(self.section)
        link = self.config.get_section(self.link)
        target = section.get(self.key)

        if isinstance(target, list):
            target.remove(self)
            target.append(link)
            section.set(self.key, target)
        else:
            section.set(self.key, link)

        # Preserve the promise for writing back out.
        section.set("_%s_promise" % (self.key), self)
        self.__fulfilled = True


class Configuration(object):
    """ Configuration class performs the loading, saving, and parsing
        of an INI-style configuration file with a few advanced features
        such as value typing, file inclusion, section references, and
        JSON-style list definition.
    """

    def __init__(self):
        """ initialise the container
            store in key:value format withing the certain category
        """

        self.__container = ConfigurationSection()

        self._filename = None

        self.loaded = False

    def __resolve_links(self):
        """ resolves all linked references (SectionPromise instances).
        """

        for promise in SectionPromise.promises:
            promise.resolve()

        SectionPromise.promises = []

    def add_section(self, section_name):
        """ Adds a new configuration section to the main dictionary.
        """

        section = ConfigurationSection()
        self.__container.set(section_name, section)

        return section

    def remove_section(self, section_name):
        """ Removes a section from the main dictionary.
        """

        del self.__container[section_name]

    @property
    def sections(self):
        """ Returns a list of all sections in the configuration.
        """

        return self.__container.keys()

    def has_section(self, section_name):
        """ Return if this configuration has a section named
            :param section_name:.
        """

        return section_name in self.__container

    def get_section(self, section_name):
        """ Return the internal ConfigurationSection representation of a
            set of configuration entries.

            :param str section_name: Section name to retrieve.
            :rtype: malibu.config.configuration.ConfigurationSection
            :return: ConfigurationSection or None
        """

        if self.__container.__contains__(section_name):
            return self.__container[section_name]
        else:
            return None

    def unload(self):
        """ Unload an entire configuration
        """

        self.__container.clear()
        self.loaded = False

    def reload(self):
        """ Reload the configuration from the initially specified file
        """

        self.unload()
        self.load(self._filename)

    def save(self, filename=None):
        """ Write the loaded configuration into the file specified by
            :param filename: or to the initially specified filename.

            All linked sections are flattened into SectionPromise instances
            and written to the configuration properly.

            :raises ValueError: if no save filename available.
        """

        if filename is None:
            filename = self._filename

        if filename is None:
            raise ValueError('No filename specified and no stored filename.')

        with closing(io.open(filename, 'w')) as config:
            for section, smap in self.__container.items():
                config.write("[%s]\n" % (section))
                for key, value in smap.items():
                    if isinstance(value, list):
                        value = "+list:" + json.dumps(value)
                    elif isinstance(value, ConfigurationSection):
                        if "_%s_promise" % (key) in smap:
                            value = str(smap["_%s_promise" % (key)])
                        else:
                            value = str(value)
                    elif isinstance(value, SectionPromise):
                        continue
                    elif isinstance(value, io.TextIOBase):
                        value = "+file:" + value.name
                    else:
                        value = str(value)
                    config.write("%s = %s\n" % (key, value))
                config.write("\n")

    def load(self, filename):
        """ Loads a INI-style configuration from the given filename.
            If the file can not be opened from :param filename:,
            a ValueError is raised. Upon any other error,
            the exception is simply raised to the top.

            :raises ValueError: if no filename provided.
            :raises Exception: upon other error
        """

        try:
            fobj = io.open(filename, 'r')
            self._filename = filename
            self.load_file(fobj)
            fobj.close()
        except IOError:
            raise ValueError("Invalid filename '%s'." % (filename))
        except:
            raise

    def load_file(self, fobj):
        """ Performs the full load of the configuration file from
            the underlying file object. If a file object is not passed in
            :param fobj:, TypeError is raised.

            :raises TypeError: if :param fobj: is not a file type
        """

        if not fobj or not isinstance(fobj, io.IOBase):
            raise TypeError("Invalid file object.")

        if self.loaded:
            self.__container.clear()

        section_name = None
        option_key = None
        option_value = None

        for line in fobj.readlines():
            line = line.strip('\n').lstrip()

            if line.startswith('#') or line.startswith(';'):
                continue
            elif line.startswith('[') and line.endswith(']'):
                # This is the beginning of a section tag.
                section_name = line[1:-1]
                if not self.get_section(section_name):
                    self.add_section(section_name)
                continue
            elif '=' in line:
                s = line.split('=', 1)
                # strip whitespace
                option_key = s[0].strip()
                option_value = s[1].strip() if s[1] is not '' else None

                if option_value and option_value[-1] == ';':
                    option_value = option_value[0:-1]
                section = self.get_section(section_name)

                if not option_value:
                    section.set(option_key, option_value)
                    continue

                if option_value.startswith('+'):  # typed reference / variable
                    dobj_type = option_value.split(':')[0][1:]
                    if len(option_value.split(':')) > 2:
                        dobj_value = ':'.join(option_value.split(':')[1:])
                    else:
                        dobj_value = option_value.split(':')[1]

                    if dobj_type.lower() == 'file':
                        try:
                            section.set(option_key, io.open(dobj_value, 'r'))
                        except:
                            try:
                                section.set(option_key, io.open(dobj_value, 'w+'))
                            except:
                                section.set(option_key, None)
                    elif dobj_type.lower() in ["url", "uri"]:
                        try:
                            section.set(option_key, urlopen(dobj_value).read())
                        except:
                            raise
                        section.set(option_key, None)
                    elif dobj_type.lower() == 'list':
                        try:
                            dobj_list = json.loads('%s' % (dobj_value))
                            dobj_list = unicode2str(dobj_list)
                        except:
                            dobj_list = []
                        dobj_repl = []
                        for item in dobj_list:
                            if item.startswith(b'@'):
                                link_name = item[1:]
                                if not self.get_section(link_name):
                                    dobj_repl.append(
                                        SectionPromise(
                                            self,
                                            section_name,
                                            option_key,
                                            link_name))
                                else:
                                    link = self.get_section(link_name)
                                    dobj_repl.append(link)
                            else:
                                dobj_repl.append(item)
                        section.set(option_key, dobj_repl)
                    else:
                        section.set(option_key, option_value)
                elif option_value.startswith('@'):  # section reference
                    link_name = option_value[1:]
                    if not self.get_section(link_name):
                        section.set(
                            option_key,
                            SectionPromise(
                                self,
                                section_name,
                                option_key,
                                link_name))
                    else:
                        link = self.get_section(link_name)
                        section.set(option_key, link)
                else:
                    section.set(option_key, option_value)
                continue
            else:
                continue

        self.__resolve_links()
        self.loaded = True



# -*- coding: utf-8 -*-
import sys


class ArgumentParser(object):

    OPTION_SINGLE = 1
    OPTION_PARAMETERIZED = 2

    PARAM_SHORT = 1
    PARAM_LONG = 2

    @classmethod
    def from_argv(cls):
        """
            Creates an ArgumentParser engine with args from
            sys.argv[].
        """
        return cls(sys.argv)

    def __init__(self, args, mapping={}):
        """ Initializes an ArgumentParser instance.  Parses out
            things that we already know, like args[0] (should be the
            executable script).
        """

        try:
            self.exec_file = args[0]
        except:
            self.exec_file = ''

        self.__args = args

        self._default_types = {
            ArgumentParser.PARAM_SHORT: ArgumentParser.OPTION_SINGLE,
            ArgumentParser.PARAM_LONG: ArgumentParser.OPTION_SINGLE
        }
        self._opt_types = {}
        self._mapping = mapping
        self._descriptions = {}
        self.options = {}
        self.parameters = []

    def __enter__(self):

        return self

    def __exit__(self, exc_type, exc_val, traceback):

        return None

    def param_defined(self, param):
        """ Checks all of the various ways a parameter can be a defined to see
            if the given parameter actually is defined.

            :param str param: Parameter to check for
            :return: parameter is defined
            :rtype: bool
        """

        if param.startswith('-'):
            param = param.lstrip('-')

        is_typed = param in self._opt_types
        is_mapped = param in self._mapping
        is_described = False

        for _p in self.get_option_descriptions().keys():
            if param in _p:
                is_described = True
            continue

        return is_typed or is_mapped or is_described

    def set_default_param_type(self, param_type, opt=OPTION_SINGLE):
        """ Sets the default type map that a parameter will be treated as.
            Can help force more uniform arguments without having to pre-define
            options.

            :param int param_type: Parameter type (PARAM_LONG, PARAM_SHORT)
            :param int opt: Option type (OPTION_PARAMETERIZED, OPTION_SINGLE)
            :return: none
            :rtype: None
        """

        self._default_types[param_type] = opt

    def add_option_type(self, option, opt=OPTION_SINGLE):
        """ Adds a type mapping to a specific option. Allowed types are
            OPTION_SINGLE and OPTION_PARAMETERIZED.

            :param str option: Option to set type for
            :param int opt: Option type (OPTION_PARAMETERIZED, OPTION_SINGLE)
            :return: none
            :rtype: None
        """

        self._opt_types[option] = opt

    def add_option_mapping(self, option, map_name):
        """ Maps a option value (-a, -g, --thing, etc.) to a full word or
            phrase that will be stored in the options dictionary after parsing
            is finished.  Makes option usage easier.

            :param str option: Option to map from
            :param str map_name: Option name to map to
            :return: none
            :rtype: None
        """

        self._mapping[option] = map_name

    def add_option_description(self, option, description):
        """ Adds a helpful description for an argument. Is returned when
            get_option_descriptions is called.

            :param str option: Option to describe
            :param str description: Description of option
            :return: none
            :rtype: None
        """

        self._descriptions[option] = description

    def get_option_descriptions(self):
        """ Returns a map of options to their respective descriptions.  Good
            for printing out a series of help messages describing usage.

            :return: Dictionary of option => description pairs
            :rtype: dict
        """

        processed_descriptions = {}
        for option, description in self._descriptions.items():
            if len(option) == 1:  # This is a flag. Prepend a dash.
                option = '-' + option
                processed_descriptions.update({option: description})
            elif len(option) > 1 and option.startswith('--'):
                # Option. Append blindly.
                processed_descriptions.update({option: description})
            elif len(option) > 1 and not option.startswith('--'):
                # Option. Append double-dash.
                option = '--' + option
                processed_descriptions.update({option: description})
            else:
                # What is this? Blindly append.
                processed_descriptions.update({option: description})

        return processed_descriptions

    def parse(self):
        """ Parses out the args into the mappings provided in self.mapping,
            stores the result in self.options.

            :return: none
            :rtype: None
        """

        waiting_argument = None

        for param in self.__args:
            if param[0] in ['"', "'"]:
                param = param.lstrip(param[0])

            if param[-1] in ['"', "'"]:
                param = param.rstrip(param[-1])

            if param.startswith('-') and not self.param_defined(param):
                # Starts with a -, but is not an option..is a parameter.
                if waiting_argument is not None:
                    self.options.update({waiting_argument: param})
                    waiting_argument = None
                else:
                    self.parameters.append(param)

                continue

            if param.startswith('--'):
                # Long option mapping.
                paraml = param[2:]
                if '=' in paraml:
                    param, value = paraml.split('=')
                    # Store a PARAMETERIZED option type, if none exists.
                    if param not in self._opt_types:
                        self.add_option_type(
                            param,
                            ArgumentParser.OPTION_PARAMETERIZED)
                else:
                    param = paraml
                    if param not in self._opt_types:
                        self.add_option_type(
                            param,
                            self._default_types[ArgumentParser.PARAM_LONG])
                store_as = param
                if param in self._mapping:
                    self.options.update({self._mapping[param]: True})
                    store_as = self._mapping[param]
                else:
                    self.options.update({param: True})
                if param in self._opt_types:
                    ptype = self._opt_types[param]
                    if ptype == ArgumentParser.OPTION_PARAMETERIZED:
                        if '=' in paraml:
                            # Option is long and contains the value.
                            self.options.update({store_as: value})
                            waiting_argument = None
                        else:
                            waiting_argument = store_as
            elif param.startswith('-'):
                # Short option mapping.
                param = param[1:]
                for opt in param:
                    store_as = opt
                    if opt in self._mapping:
                        self.options.update({self._mapping[opt]: True})
                        store_as = self._mapping[opt]
                    else:
                        self.options.update({opt: True})
                    if opt in self._opt_types:
                        ptype = self._opt_types[opt]
                        if ptype == ArgumentParser.OPTION_PARAMETERIZED:
                            waiting_argument = store_as
            else:
                # Option parameter or command parameter.
                if waiting_argument is not None:
                    self.options.update({waiting_argument: param})
                    waiting_argument = None
                else:
                    self.parameters.append(param)



# -*- coding: utf-8 -*-
import glob
import os

modules = glob.glob(os.path.dirname(__file__) + "/*.py")
__all__ = [os.path.basename(f)[:-3] for f in modules
           if not os.path.basename(f).startswith('_') and
           not f.endswith('__init__.py') and os.path.isfile(f)]

__doc__ = """
Relatively simplistic configuration file loader, parser, and writer.

Has the ability to cross-link sections and load into completely de-serialized
Python dictionaries.

Example of configuration format:

.. code-block:: ini

    ; filename: example.ini

    [server]
    address = 127.0.0.1
    port = 1234

    [default_route]
    uri = /
    controller = example_project.views.ExampleView

    [static_content]
    uri = /other_place
    content = +url:https://maio.me/~pirogoeth/VIRUS.txt

    [file_content]
    uri = /another_place
    content = +file:/var/data/content.txt

    [routes]
    routes = +list:["@default_route"]

Usage:

.. code-block:: python

    from malibu.config import configuration
    conf = configuration.Configuration()
    conf.load("example.ini")

    # Grab a section
    srv_opts = conf.get_section("server")

    # Iterate over sections
    for section in conf.sections:
        # do section...
        pass

    # Add a section
    data = {
        "value": "abc",
        "num": 123
    }

    conf.add_section("test", data)
    conf.add_section("test2", data)

    # Remove a section
    conf.remove_section("test2")

    # Save the changed configuration
    conf.save()

    # Or, save to a new file
    conf.save(filename = "/var/data/config2.ini")

.. automodule:: malibu.config.configuration
   :members:

"""



# -*- coding: utf-8 -*-
import glob
import os


modules = glob.glob(os.path.dirname(__file__) + "/*.py")
__all__ = [os.path.basename(f)[:-3] for f in modules
           if not os.path.basename(f).startswith('_') and
           not f.endswith('__init__.py') and os.path.isfile(f)]
